import json
from enum import Enum
from typing import List


# 基础大类的名称
class Category0(str, Enum):
    Halide = "Halide"
    Amine = "Amine"
    Alcohol = "Alcohol"
    Boronic_Acid = "Boronic Acid"
    Carboxylic_Acid = "Carboxylic Acid"
    Acid_Chloride = "Acid Chloride"
    Sulfonyl_Chloride = "Sulfonyl Chloride"
    Aldehyde = "Aldehyde"
    Ketone = "Ketone"
    Default = "-"
    Error = "Category Error"
    # Alkyne = "Alkyne"
    # Olefin = "Olefin"
    # Epoxides = "Epoxides"
    # Hydrazines = "Hydrazines"
    # Isocyanates = "Isocyanates"
    # Isothiocyanates = "Isothiocyanates"
    # Nitriles = "Nitriles"
    # Ester = "Ester"
    Others = "Others"


# 基础大类含有的官能团，可根据验证结果修改SMARTS
class Category0FunctionalGroup(str, Enum):
    Halide = "[F,Cl,Br,I]"
    Halide_I = "[I]"
    Halide_Br = "[Br]"
    Halide_Cl = "[Cl]"
    Halide_F = "[F]"
    Amine = "[NH2,NH,N,nH]"
    Primary_or_Benzyl_Amine = '[NH2]'
    Secondary_Amine = '[NH]'
    Secondary_Amine_R = '[nH]'
    Tertiary_Amine = '[N]'
    Alcohol = "[OH]"
    Boronic_Acid = "B(O)O"
    Carboxylic_Acid = "C(=O)O"
    Acid_Chloride = "C(=O)Cl"
    Sulfonyl_Chloride = "S(=O)(Cl)=O"
    Aldehyde = "[CH]=O"
    Ketone = "[CX3;$(C([#6])(=[O])[#6])]"
    # Alkyne = "C#C"
    # Olefin = "C=C"
    # Epoxides = "C1CO1"
    # Hydrazines = "NN"
    # Isocyanates = "N=C=O"
    # Isothiocyanates = "N=C=S"
    # Nitriles = "C#N"
    # Ester1 = "COC=O"
    # Ester2 = "CCOC=O"
    # Ester3 = "cOC=O"
    # Ester4 = "cCOC=O"
    # Ester5 = "CcOC=O"
    # Ester6 = "ccOC=O"


# Category4的特定子结构的SMARTS
class Category4Smarts(str, Enum):
    # 5元杂环
    Pyrrole_3H = "N1=CCC=C1"
    Pyrrole_2H = "N1CC=CC=1"
    Pyrrole_2H_Enol = "N1CC=cc=1"
    Pyrazole_3H = "N1=NCC=C1"
    Imidazole_4H = "N1C=NCC=1"
    Triazole_124_3H = "N1=NCN=C1"
    Triazole_123_4H = "N1N=NCC=1"
    Tetrazole_5H = "N1=NN=NC1"

    Furan = "o1cccc1"  # 呋喃
    Oxazole = "o1cncc1"  # 恶唑
    Isoxazole = "o1nccc1"  # 异恶唑
    Oxadiazole_123 = "o1nncc1"
    Oxadiazole_124 = "o1ncnc1"
    Oxatriazole_1235 = "o1nncn1"
    Oxatriazole_1234 = "o1nnnc1"
    Tetranitrogen_monooxide = "o1nnnn1"

    Thiophene = "s1cccc1"  # 噻吩
    Thiazole = "s1cncc1"  # 噻唑
    Isothiazole = "s1nccc1"
    Thiadiazole_123 = "s1nncc1"
    Thiadiazole_124 = "s1ncnc1"
    Thiadiazole_134 = "s1cnnc1"
    Thiatriazole_1235 = "s1nncn1"
    Thiatriazole_1234 = "s1nnnc1"
    Tetranitrogen_monosulfide = "s1nnnn1"

    Pyrrole = "[nH,n]1cccc1"  # 吡咯
    Imidazole_nH = "[nH]1cncc1"  # 13唑 Imidazole  1,3-Diazole
    Imidazole_n = "[n]1cncc1"  # 13唑 Imidazole  1,3-Diazole
    Pyrazole = "[nH,nX3]1nccc1"  # 12唑 Pyrazole  1,2-Diazole
    Triazole_123 = "[nH,n]1nncc1"  # 123三唑
    Triazole_124 = "[nH,n]1ncnc1"  # 124三唑
    Tetrazole_1235 = "[nH,n]1nncn1"
    Tetrazole_1234 = "[nH,n]1nnnc1"
    Pentazole_1H = "[nH,n]1nnnn1"

    # 6元杂环
    Pyridine = "n1ccccc1"  # 吡啶
    Pyridazine = "n1ncccc1"  # 哒嗪
    Pyrimidine = "n1cnccc1"  # 嘧啶
    Pyrazine = "n1ccncc1"  # 吡嗪
    Triazine_123 = "n1nnccc1"
    Triazine_124 = "n1ncncc1"
    Triazine_135 = "n1cncnc1"
    Tetrazine_1234 = "n1nnncc1"
    Tetrazine_1245 = "n1ncnnc1"
    Tetrazine_1235 = "n1cnnnc1"
    Pentazine = "n1nnnnc1"
    Hexazine = "n1nnnnn1"


class Category1(str, Enum):
    Aromatic = "Aromatic"
    Aliphatic = "Aliphatic"
    Free_Hydrogen = "Free Hydrogen"
    Enol = "Enol"  # 烯醇
    Default = "-"
    Error = "Category1 Error"


class Category2(str, Enum):
    Default = "-"
    Error = "Category2 Error"


class Category2Aromatic(str, Enum):
    Benzene = "Benzene"
    Heterocyclic = "Heterocyclic"
    Fused_Ring = "Fused Ring"
    Quinone = "Quinone"
    Default = "-"
    Error = "Category2 Error"


class Category2Aliphatic(str, Enum):
    Primary = "Primary"
    Secondary = "Secondary"
    Tertiary = "Tertiary"
    Benzyl = "Benzyl"
    Default = "-"
    Error = "Category2 Error"


class Category3(str, Enum):
    FiveMembered = "Five Membered"
    SixMembered = "Six Membered"
    Default = "-"
    Error = "Category3 Error"


class Category4(str, Enum):
    Default = "-"
    Error = "Category4 Error"


class Category4FiveMembered(str, Enum):
    # 5元杂环 # 30
    Pyrrole_3H = "3H-Pyrrole"
    Pyrrole_2H = "2H-Pyrrole"
    Pyrazole_3H = "3H-Pyrazole"
    Imidazole_4H = "4H-Imidazole"
    Triazole_124_3H = "3H-1,2,4-Triazole"
    Triazole_123_4H = "4H-1,2,3-Triazole"
    Tetrazole_5H = "5H-Tetrazole"

    Furan = "Furan"  # 呋喃
    Oxazole = "Oxazole"  # 恶唑
    Isoxazole = "Isoxazole"  # 异恶唑
    Oxadiazole_123 = "1,2,3-Oxadiazole"
    Oxadiazole_124 = "1,2,4-Oxadiazole"
    Oxatriazole_1235 = "1,2,3,5-Oxatriazole"
    Oxatriazole_1234 = "1,2,3,4-Oxatriazole"
    Tetranitrogen_monooxide = "Tetranitrogen(III) monooxide"

    Thiophene = "Thiophene"  # 噻吩
    Thiazole = "Thiazole"  # 噻唑
    Isothiazole = "Isothiazole"
    Thiadiazole_123 = "1,2,3-Thiadiazole"
    Thiadiazole_124 = "1,2,4-Thiadiazole"
    Thiadiazole_134 = "1,3,4-Thiadiazole"
    Thiatriazole_1235 = "1,2,3,5-Thiatriazole"
    Thiatriazole_1234 = "1,2,3,4-Thiatriazole"
    Tetranitrogen_monosulfide = "Tetranitrogen(III) monosulfide"

    Pyrrole = "Pyrrole"  # 吡咯
    Imidazole = "Imidazole"  # 13唑 Imidazole  1,3-Diazole
    Pyrazole = "Pyrazole"  # 12唑 Pyrazole  1,2-Diazole
    Triazole_123 = "1,2,3-Triazole"  # 123三唑
    Triazole_124 = "1,2,4-Triazole"  # 124三唑
    Tetrazole = "Tetrazole"
    Pentazole_1H = "1H-Pentazole"
    Default = "-"


class Category4SixMembered(str, Enum):
    # 6元杂环 # 12
    Pyridine = "Pyridine"  # 吡啶
    Pyridazine = "Pyridazine"  # 哒嗪
    Pyrimidine = "Pyrimidine"  # 嘧啶
    Pyrazine = "Pyrazine"  # 吡嗪
    Triazine_123 = "1,2,3-Triazine"
    Triazine_124 = "1,2,4-Triazine"
    Triazine_135 = "1,3,5-Triazine"
    Tetrazine_1234 = "1,2,3,4-Tetrazine"
    Tetrazine_1245 = "1,2,4,5-Tetrazine"
    Tetrazine_1235 = "1,2,3,5-Tetrazine"
    Pentazine = "Pentazine"
    Hexazine = "Hexazine"


class Category5(int, Enum):
    Pos_1 = 1
    Pos_2 = 2
    Pos_3 = 3
    Pos_4 = 4
    Pos_5 = 5
    Pos_6 = 6


class Hydrolysis(str, Enum):
    Yes = "Y"
    No = "N"
    Default = "-"


class Steric_Hindrance(str, Enum):
    Yes = "Y"
    No = "N"
    Default = "-"


class CategoryResult:
    def __init__(self, category0: Category0, category1: Category1,
                 category2: Category2, category3: Category3,
                 category4: Category4, category5: str,
                 hydrolysis: Hydrolysis, steric_hindrance: Steric_Hindrance):
        self.category0 = category0
        self.category1 = category1
        self.category2 = category2
        self.category3 = category3
        self.category4 = category4
        self.category5 = category5
        self.hydrolysis = hydrolysis
        self.steric_hindrance = steric_hindrance

    def toJson(self):
        return json.dumps(self, default=lambda o: o.__dict__,
                          sort_keys=True, indent=4)



from rdkit import Chem
from rdkit.Chem import rdqueries
from collections.abc import Iterable
from classification.models import Category0, Category5, Category2Aliphatic, Category4Smarts
from classification.models import Category0FunctionalGroup as C0FG


def mol_with_atom_index(mol):
    for atom in mol.GetAtoms():
        atom.SetAtomMapNum(atom.GetIdx())
    return mol


# 类型转换，把输入的category0:string 转换为Category0，以便最后的输出
def convert_category_0(category0: str) -> Category0:
    if category0 == Category0.Halide.value:
        return Category0.Halide
    elif category0 == Category0.Amine.value:
        return Category0.Amine
    elif category0 == Category0.Alcohol.value:
        return Category0.Alcohol
    elif category0 == Category0.Boronic_Acid.value:
        return Category0.Boronic_Acid
    elif category0 == Category0.Carboxylic_Acid.value:
        return Category0.Carboxylic_Acid
    elif category0 == Category0.Acid_Chloride.value:
        return Category0.Acid_Chloride
    elif category0 == Category0.Sulfonyl_Chloride.value:
        return Category0.Sulfonyl_Chloride
    elif category0 == Category0.Aldehyde.value:
        return Category0.Aldehyde
    elif category0 == Category0.Ketone.value:
        return Category0.Ketone
    # elif category0 == Category0.Alkyne.value:
    #     return Category0.Alkyne
    # elif category0 == Category0.Olefin.value:
    #     return Category0.Olefin
    # elif category0 == Category0.Epoxides.value:
    #     return Category0.Epoxides
    # elif category0 == Category0.Hydrazines.value:
    #     return Category0.Hydrazines
    # elif category0 == Category0.Isocyanates.value:
    #     return Category0.Isocyanates
    # elif category0 == Category0.Isothiocyanates.value:
    #     return Category0.Isothiocyanates
    # elif category0 == Category0.Nitriles.value:
    #     return Category0.Nitriles
    # elif category0 == Category0.Ester.value:
    #     return Category0.Ester
    else:
        return Category0.Error


# 根据category0匹配其对应的原子索引和官能团
def functional_group_from_category0(smile, category0):
    if category0 == Category0.Halide.value:
        # 脂肪类优先于芳香类，优先级按照I Br Cl F，脂肪类F不考虑
        if identify_functional_group(smile, 'C[I]')[1]:
            return identify_functional_group(smile, '[I]')[0], C0FG.Halide_I.value
        elif identify_functional_group(smile, 'C[Br]')[1]:
            return identify_functional_group(smile, '[Br]')[0], C0FG.Halide_Br.value
        elif identify_functional_group(smile, 'C[Cl]')[1]:
            return identify_functional_group(smile, '[Cl]')[0], C0FG.Halide_Cl.value
        elif identify_functional_group(smile, 'c[I]')[1]:
            return identify_functional_group(smile, '[I]')[0], C0FG.Halide_I.value
        elif identify_functional_group(smile, 'c[Br]')[1]:
            return identify_functional_group(smile, '[Br]')[0], C0FG.Halide_Br.value
        elif identify_functional_group(smile, 'c[Cl]')[1]:
            return identify_functional_group(smile, '[Cl]')[0], C0FG.Halide_Cl.value
        elif identify_functional_group(smile, 'c[F]')[1]:
            return identify_functional_group(smile, '[F]')[0], C0FG.Halide_F.value
        elif identify_functional_group(smile, 'C(F)')[1]:
            return identify_functional_group(smile, '[F]')[0], C0FG.Halide_F.value
        elif identify_functional_group(smile, '[n+]I')[1]:
            return identify_functional_group(smile, '[I]')[0], C0FG.Halide_I.value
        elif identify_functional_group(smile, '[n+]F')[1]:
            return identify_functional_group(smile, '[F]')[0], C0FG.Halide_F.value
        else:
            raise Exception('Incorrect Halide classification')
    elif category0 == Category0.Amine.value:
        # Amine优先芳香性alpha-c
        if identify_functional_group(smile, '[nH]')[1] and identify_functional_group(smile, 'c[nH]')[1] \
                or identify_functional_group(smile, 'C[nH]')[1]:
            return identify_functional_group(smile, '[nH]')[0], C0FG.Secondary_Amine_R.value
        elif identify_functional_group(smile, '[NH2]')[1] and identify_functional_group(smile, 'c[NH2]')[1] \
                or identify_functional_group(smile, 'C[NH2]')[1]:
            return identify_functional_group(smile, '[NH2]')[0], C0FG.Primary_or_Benzyl_Amine.value
        elif identify_functional_group(smile, '[NH]')[1] and identify_functional_group(smile, 'c[NH]')[1] \
                or identify_functional_group(smile, 'C[NH]')[1]:
            return identify_functional_group(smile, '[NH]')[0], C0FG.Secondary_Amine.value
        elif identify_functional_group(smile, '[N]')[1] and identify_functional_group(smile, 'c[N]')[1] \
                or identify_functional_group(smile, 'C[N]')[1]:
            return identify_functional_group(smile, '[N]')[0], C0FG.Tertiary_Amine.value
        else:
            raise Exception('Incorrect Amine classification')
    elif category0 == Category0.Alcohol.value:
        (matches, flag) = identify_functional_group(smile, C0FG.Alcohol.value)
        return matches, C0FG.Alcohol.value
    elif category0 == Category0.Boronic_Acid.value:
        (matches, flag) = identify_functional_group(smile, C0FG.Boronic_Acid.value)
        return matches, C0FG.Boronic_Acid.value
    elif category0 == Category0.Carboxylic_Acid.value:
        (matches, flag) = identify_functional_group(smile, C0FG.Carboxylic_Acid.value)
        return matches, C0FG.Carboxylic_Acid.value
    elif category0 == Category0.Acid_Chloride.value:
        (matches, flag) = identify_functional_group(smile, C0FG.Acid_Chloride.value)
        return matches, C0FG.Acid_Chloride.value
    elif category0 == Category0.Sulfonyl_Chloride.value:
        (matches, flag) = identify_functional_group(smile, C0FG.Sulfonyl_Chloride.value)
        return matches, C0FG.Sulfonyl_Chloride.value
    elif category0 == Category0.Aldehyde.value:
        (matches, flag) = identify_functional_group(smile, C0FG.Aldehyde.value)
        return matches, C0FG.Aldehyde.value
    elif category0 == Category0.Ketone.value:
        (matches, flag) = identify_functional_group(smile, C0FG.Ketone.value)
        return matches, C0FG.Ketone.value
    # elif category0 == Category0.Alkyne.value:
    #     (matches, flag) = identify_functional_group(smile, C0FG.Alkyne.value)
    #     return matches, C0FG.Alkyne.value
    # elif category0 == Category0.Olefin.value:
    #     (matches, flag) = identify_functional_group(smile, C0FG.Olefin.value)
    #     return matches, C0FG.Olefin.value
    # elif category0 == Category0.Epoxides.value:
    #     (matches, flag) = identify_functional_group(smile, C0FG.Epoxides.value)
    #     return matches, C0FG.Epoxides.value
    # elif category0 == Category0.Hydrazines.value:
    #     (matches, flag) = identify_functional_group(smile, C0FG.Hydrazines.value)
    #     return matches, C0FG.Hydrazines.value
    # elif category0 == Category0.Isocyanates.value:
    #     (matches, flag) = identify_functional_group(smile, C0FG.Isocyanates.value)
    #     return matches, C0FG.Isocyanates.value
    # elif category0 == Category0.Isothiocyanates.value:
    #     (matches, flag) = identify_functional_group(smile, C0FG.Isothiocyanates.value)
    #     return matches, C0FG.Isothiocyanates.value
    # elif category0 == Category0.Nitriles.value:
    #     (matches, flag) = identify_functional_group(smile, C0FG.Nitriles.value)
    #     return matches, C0FG.Nitriles.value
    # elif category0 == Category0.Ester.value:
    #     if identify_functional_group(smile, C0FG.Ester1.value)[1]:
    #         return identify_functional_group(smile, C0FG.Ester1.value)[0], C0FG.Ester1.value
    #     elif identify_functional_group(smile, C0FG.Ester2.value)[1]:
    #         return identify_functional_group(smile, C0FG.Ester2.value)[0], C0FG.Ester2.value
    #     elif identify_functional_group(smile, C0FG.Ester3.value)[1]:
    #         return identify_functional_group(smile, C0FG.Ester3.value)[0], C0FG.Ester3.value
    #     elif identify_functional_group(smile, C0FG.Ester4.value)[1]:
    #         return identify_functional_group(smile, C0FG.Ester4.value)[0], C0FG.Ester4.value
    #     elif identify_functional_group(smile, C0FG.Ester5.value)[1]:
    #         return identify_functional_group(smile, C0FG.Ester5.value)[0], C0FG.Ester5.value
    #     elif identify_functional_group(smile, C0FG.Ester6.value)[1]:
    #         return identify_functional_group(smile, C0FG.Ester6.value)[0], C0FG.Ester6.value
    #     else:
    #         raise Exception('Incorrect Ester classification')
    else:
        raise Exception('Incorrect compound classification')


# 根据category0反推alpha c index
def alpha_c_index_by_category(smile, category0):
    # Halide大类先脂肪后芳香
    if category0 == Category0.Halide.value:
        is_al, match_al = is_aliphatic_carbon(smile, category0)
        if is_al:
            return match_al
        else:
            is_ar, match_ar = is_aromatic_carbon(smile, category0)
            return match_ar
    # 醇类要去掉羧基的干扰
    elif category0 == Category0.Alcohol.value:
        if identify_functional_group(smile, 'C(=O)O')[1]:
            # 羧基中的C索引
            ex_c_idx = identify_functional_group(smile, 'C(=O)O')[0][0][0]
            is_ar, match_ar = is_aromatic_carbon(smile, category0)
            if is_ar:
                return delete_all_x_idx(match_ar, ex_c_idx)
            else:
                is_al, match_al = is_aliphatic_carbon(smile, category0)
                return delete_all_x_idx(match_al, ex_c_idx)
        else:
            is_ar, match_ar = is_aromatic_carbon(smile, category0)
            if is_ar:
                return match_ar
            else:
                is_al, match_al = is_aliphatic_carbon(smile, category0)
                return match_al
    # 其余大类先芳香后脂肪
    else:
        is_ar, match_ar = is_aromatic_carbon(smile, category0)
        if is_ar:
            return match_ar
        else:
            is_al, match_al = is_aliphatic_carbon(smile, category0)
            return match_al


# 判断与芳香碳的匹配
def is_aromatic_carbon(smile, category0):
    res = functional_group_from_category0(smile, category0)
    matches_ar = identify_functional_group(smile, 'c' + res[1])
    if not matches_ar[1]:
        matches_ar = identify_functional_group(smile, 'n' + res[1])
    return matches_ar[1], list(set([i[0] for i in matches_ar[0]]))


# 判断与脂肪C的匹配
def is_aliphatic_carbon(smile, category0):
    res = functional_group_from_category0(smile, category0)
    matches_al = identify_functional_group(smile, 'C' + res[1])
    return matches_al[1], list(set([i[0] for i in matches_al[0]]))


# 验证化合物是否含有官能团
def identify_functional_group(smile, smart):
    mol = Chem.MolFromSmiles(smile)
    functional_group = Chem.MolFromSmarts(smart)
    matches = mol.GetSubstructMatches(functional_group)
    return matches, len(matches) > 0


# 验证化合物是否为活泼氢
def is_free_H(smile, category0):
    # 不含活泼氢官能团（OH COOH NH2 NH）的大类，直接匹配活泼氢官能团子结构
    if category0 == Category0.Halide.value or category0 == Category0.Acid_Chloride.value or \
            category0 == Category0.Sulfonyl_Chloride.value or category0 == Category0.Aldehyde.value or \
            category0 == Category0.Ketone.value:  # or category0 == Category0.Alkyne.value or category0 == Category0.Olefin.value or category0 == Category0.Epoxides.value or category0 == Category0.Isocyanates.value or category0 == Category0.Isothiocyanates.value or category0 == Category0.Nitriles.value or category0 == Category0.Ester.value:
        matches_h = identify_functional_group(smile, '[OH,NH2,NH,nH]')
        matches_cooh = identify_functional_group(smile, 'C(=O)[OH1]')
        return len(matches_h[0]) > 0 or len(matches_cooh[0]) > 0
    # 胺类直接匹配OH COOH
    elif category0 == Category0.Amine.value:  # or category0 == Category0.Hydrazines.value:
        # 优先匹配[NH2]
        matches_nh2 = identify_functional_group(smile, C0FG.Primary_or_Benzyl_Amine.value)
        # 含有1个[NH2]以上则为活泼氢
        if len(matches_nh2[0]) > 1:
            return True
        # 恰有1个[NH2], 再看是否含有[NH]或[nH]或[OH,COOH]
        elif len(matches_nh2[0]) == 1:
            matches_nh = identify_functional_group(smile, C0FG.Secondary_Amine.value)
            matches_nhr = identify_functional_group(smile, C0FG.Secondary_Amine_R.value)
            matches_h = identify_functional_group(smile, '[OH]')
            matches_cooh = identify_functional_group(smile, 'C(=O)[OH1]')
            return len(matches_nh[0]) > 0 or len(matches_nhr[0]) > 0 or len(matches_h[0]) > 0 or len(
                matches_cooh[0]) > 0
        # 不含[NH2]
        else:
            # 看[NH]或[nH]的个数
            matches_nh = identify_functional_group(smile, C0FG.Secondary_Amine.value)
            matches_nhr = identify_functional_group(smile, C0FG.Secondary_Amine_R.value)
            if len(matches_nh[0]) > 1 or len(matches_nhr[0]) > 1 or (len(matches_nh[0]) + len(matches_nhr[0])) > 1:
                return True
            # elif len(matches_nh[0]) == 1 or len(matches_nhr[0]) == 1:
            #     matches_h = identify_functional_group(smile, '[OH,COOH]')
            #     return len(matches_h[0]) > 0
            else:
                matches_h = identify_functional_group(smile, '[OH]')
                matches_cooh = identify_functional_group(smile, 'C(=O)[OH1]')
                return len(matches_h[0]) > 0 or len(matches_cooh[0]) > 0

    # 醇类多元醇或含有活泼氢官能团
    elif category0 == Category0.Alcohol.value:
        # matches_al = identify_functional_group(smile, C0FG.Alcohol.value)
        matches_h = identify_functional_group(smile, '[NH2,NH,nH]')
        matches_cooh = identify_functional_group(smile, 'C(=O)[OH1]')
        return len(matches_h[0]) > 0 or len(matches_cooh[0]) > 0
    # 羧酸多元羧酸或含有活泼氢官能团
    elif category0 == Category0.Carboxylic_Acid.value:
        flag = False
        matches_ca = identify_functional_group(smile, C0FG.Carboxylic_Acid.value)
        if len(matches_ca[0]) > 1:
            flag = True
        else:
            matches_h = identify_functional_group(smile, '[OH,NH2,NH,nH]')
            for h in matches_h[0]:
                for ca in matches_ca[0]:
                    if h[0] not in ca:
                        flag = True
        return flag
    # 硼酸本身含有2个OH
    elif category0 == Category0.Boronic_Acid.value:
        flag = False
        matches_ba = identify_functional_group(smile, C0FG.Boronic_Acid.value)
        matches_h = identify_functional_group(smile, '[OH,NH2,NH,nH]')
        for h in matches_h[0]:
            for ba in matches_ba[0]:
                if h[0] not in ba:
                    flag = True
                    return flag
        matches_cooh = identify_functional_group(smile, 'C(=O)[OH1]')
        for cooh in matches_cooh[0]:
            for ba in matches_ba[0]:
                if cooh[0] not in ba:
                    flag = True
                    return flag
        return flag
    else:
        raise Exception('Incorrect compound classification')


# 验证化合物对脂肪类和环类的匹配度，即alpha-c所在的环是否为环状结构
def match_alpha_c_ali_aro(smile, func_group):
    smart_ali = 'C' + func_group
    smart_aro = '[R]' + func_group
    m = Chem.MolFromSmiles(smile)
    substructure_ali = Chem.MolFromSmarts(smart_ali)
    substructure_aro = Chem.MolFromSmarts(smart_aro)
    matches_ali = m.GetSubstructMatches(substructure_ali)
    matches_aro = m.GetSubstructMatches(substructure_aro)
    return matches_ali, matches_aro


# 验证alpha-c所在的环是否为芳香环
def is_ring_aromatic(smile, category0):
    m = Chem.MolFromSmiles(smile)
    # alpha-c原子索引列表
    ac = alpha_c_index_by_category(smile, category0)
    ac_ring = []
    for ri in m.GetRingInfo().AtomRings():
        for c_idx in ac:
            if c_idx in ri:
                ac_ring.append(ri)
    # 查询分子中所有芳香性原子的索引，结果为set集合
    q = rdqueries.IsAromaticQueryAtom()
    ar_atom = {x.GetIdx() for x in m.GetAtomsMatchingQuery(q)}
    # alpha-c所在环的碳原子索引集合 vs 分子中芳香性原子索引集合
    if set(get_item(ac_ring)) != set():
        for ac_r in ac_ring:
            if set(get_item(ac_r)).issubset(ar_atom):
                return True
        return False
    else:
        return False


# 验证某个index是否在环上以及环是否为芳香环
def is_idx_in_ring_and_aromatic(smile, index):
    m = Chem.MolFromSmiles(smile)
    # 得到环信息，并循环
    for ri in m.GetRingInfo().AtomRings():
        if index in ri:
            # index在环上，进一步判断环的芳香性
            q = rdqueries.IsAromaticQueryAtom()
            ar_atom = {x.GetIdx() for x in m.GetAtomsMatchingQuery(q)}
            # index所在环的碳原子索引集合 vs 分子中芳香性原子索引集合
            if set(get_item(ri)).issubset(ar_atom):
                return True
            else:
                return False
        else:
            return False


# 判断idx是否是两个环的并点
def is_idx_in_double_ring(smile, idx):
    m = Chem.MolFromSmiles(smile)
    ct = 0
    for ri in m.GetRingInfo().AtomRings():
        if idx in ri:
            ct += 1
    return ct


# 根据alpha-c上的H区分脂肪类
def count_H_on_alpha_c(smile, category0):
    idx_list = alpha_c_index_by_category(smile, category0)
    alpha_c_h_num = []
    for idx in idx_list:
        h_num = count_H(smile, idx)
        alpha_c_h_num.append(h_num)
    if sum(alpha_c_h_num) == 2 * len(alpha_c_h_num) or max(alpha_c_h_num) >= 2:
        return Category2Aliphatic.Primary
    elif sum(alpha_c_h_num) == len(alpha_c_h_num) or max(alpha_c_h_num) == 1:
        return Category2Aliphatic.Secondary
    elif sum(alpha_c_h_num) == 0:
        return Category2Aliphatic.Tertiary
    else:
        return Category2Aliphatic.Error


# 验证是否为并环
def is_fused_ring(smile, category0):
    flag = False
    # 先拿到alpha-c列表
    alpha_c_idx_list = alpha_c_index_by_category(smile, category0)
    # 再拿到环列表
    mol = Chem.MolFromSmiles(smile)
    ring_atom_idx_list = mol.GetRingInfo().AtomRings()
    # 定义含有alpha-c的环
    alpha_c_rings = []
    for alpha_c_idx in alpha_c_idx_list:
        for ring_atom_idx in ring_atom_idx_list:
            if alpha_c_idx in ring_atom_idx:
                alpha_c_rings.append(ring_atom_idx)
    # alpha-c环与环列表匹配，各自都必须是芳香环且交集个数等于2则是并环
    for alpha_c_ring in alpha_c_rings:
        for ring_atom_idx in ring_atom_idx_list:
            acr_is_ar = is_tuple_aromatic(smile, alpha_c_ring)
            rai_is_ar = is_tuple_aromatic(smile, ring_atom_idx)
            if alpha_c_ring != ring_atom_idx and acr_is_ar and rai_is_ar \
                    and len(set(alpha_c_ring).intersection(ring_atom_idx)) == 2:
                flag = True
    return flag


# 验证某tuple是否为芳香性环
def is_tuple_aromatic(smile, t):
    m = Chem.MolFromSmiles(smile)
    q = rdqueries.IsAromaticQueryAtom()
    ar_atom = {x.GetIdx() for x in m.GetAtomsMatchingQuery(q)}
    if len(t):
        return set(get_item(t)).issubset(ar_atom)
    else:
        return False


# 验证某tuple是否为自定义芳香性环
def is_tuple_custom_aromatic(smile, t):
    m = Chem.MolFromSmiles(smile)
    sp2_flag = 1
    for idx in t:
        if str(m.GetAtomWithIdx(idx).GetHybridization()) != 'SP2':
            sp2_flag = 0
    return sp2_flag


# 验证苯环还是杂环，前提是上一级分类为芳香环，且非并环
# 从alpha-c索引入手查是否包含在环索引元组中，然后根据个数判断5元6元
def is_ring_x_atom(smile, category0):
    alpha_c_idx_list = alpha_c_index_by_category(smile, category0)
    ring_len = []
    for alpha_c_idx in alpha_c_idx_list:
        num = count_ring_atom(smile, alpha_c_idx)
        if num:
            ring_len.append(num)

    return ring_len


# 判断alpha-c所在的环是不是杂环，前提是该化合物已经被分为芳香类
def is_alpha_ring_heterocyclic(smile, category0):
    # alpha-c列表
    alpha_c_idx_list = alpha_c_index_by_category(smile, category0)
    # 环tuple列表
    mol = Chem.MolFromSmiles(smile)
    ring_list = mol.GetRingInfo().AtomRings()
    ar_ring_list = []
    for t in ring_list:
        if is_tuple_custom_aromatic(smile, t):
            ar_ring_list.append(t)
    is_hete = 0
    for idx in alpha_c_idx_list:
        for ring in ar_ring_list:
            if idx in ring:
                for ri in ring:
                    if mol.GetAtomWithIdx(ri).GetAtomicNum() != 6:
                        is_hete = 1
    return is_hete


# 某个原子上H的个数
def count_H(smile, c_index):
    m = Chem.MolFromSmiles(smile)
    num = m.GetAtomWithIdx(c_index).GetTotalNumHs()
    return num


def is_aromatic(smile, c_index):
    m = Chem.MolFromSmiles(smile)
    return m.GetAtomWithIdx(c_index).GetIsAromatic()


# 判断某索引所在环的原子个数
def count_ring_atom(smile, alpha_c_idx):
    m = Chem.MolFromSmiles(smile)
    ri = m.GetRingInfo()
    num = 0
    for r in ri.AtomRings():
        if is_tuple_aromatic(smile, r):
            if alpha_c_idx in r:
                num = len(r)
    if num:
        return num
    else:
        for r in ri.AtomRings():
            if is_tuple_custom_aromatic(smile, r):
                if alpha_c_idx in r:
                    num = len(r)
        return num


# 判断alpha-c在环上的位置，前提是上一级为5、6元杂环
def locate_alpha_c_index_on_heterocyclic(smile, category0, heterocyclic):
    # 得到alpha-c index
    alpha_c_idx = alpha_c_index_by_category(smile, category0)
    # 匹配对应的杂环
    (matches, flag) = identify_functional_group(smile, heterocyclic)
    position_list = []
    for alpha_c in alpha_c_idx:
        for match in matches:
            if alpha_c in match:
                position = match.index(alpha_c) + 1
                # 对于对称结构做互补替换
                # 五元杂环对称结构5->2, 4->3
                if heterocyclic in [Category4Smarts.Furan.value, Category4Smarts.Tetranitrogen_monooxide.value,
                                    Category4Smarts.Thiophene.value, Category4Smarts.Thiadiazole_134.value,
                                    Category4Smarts.Tetranitrogen_monosulfide.value, Category4Smarts.Pyrrole.value,
                                    Category4Smarts.Pentazole_1H.value]:
                    if position == 5:
                        position = 2
                    elif position == 4:
                        position = 3
                # 六元杂环对称结构6->2, 5->3
                elif heterocyclic in [Category4Smarts.Pyridine.value, Category4Smarts.Hexazine.value]:
                    if position == 6:
                        position = 2
                    elif position == 5:
                        position = 3
                # 六元杂环中心对称结构
                elif heterocyclic in [Category4Smarts.Pyrazine.value]:
                    if position in [3, 5, 6]:
                        position = 2
                # 六元杂环中心对称结构
                elif heterocyclic in [Category4Smarts.Triazine_135.value]:
                    if position in [2, 4, 6]:
                        position = 2
                # 六元杂环中心对称结构
                elif heterocyclic in [Category4Smarts.Tetrazine_1235.value]:
                    if position in [2, 6]:
                        position = 2
                position_list.append(position)

    return position_list


# 根据position_list返回官能团位置信息list
def mapping_from_position_list(pos):
    pos_info = []
    for p in pos:
        if p == Category5.Pos_1.value:
            pos_info.append("1")
        elif p == Category5.Pos_2.value:
            pos_info.append("2")
        elif p == Category5.Pos_3.value:
            pos_info.append("3")
        elif p == Category5.Pos_4.value:
            pos_info.append("4")
        elif p == Category5.Pos_5.value:
            pos_info.append("5")
        elif p == Category5.Pos_6.value:
            pos_info.append("6")
        else:
            pos_info.append("-")

    return pos_info


# 判断是否易水解
def is_hydrolysis(smile, category0):
    if identify_functional_group(smile, 'C#N')[1]:  # 氰基CN
        return True
    elif identify_functional_group(smile, 'C(=O)O[CH3]')[1]:  # 甲酯COOCH3
        return True
    elif identify_functional_group(smile, 'C(=O)O[CH2][CH3]')[1]:  # 乙酯COOCH2CH3
        return True
    else:
        return False


# 查找alpha-c上邻位c的idx
def neighbor_c_idx_of_alpha_c(smile, category0):
    mol = Chem.MolFromSmiles(smile)
    ring = mol.GetRingInfo().AtomRings()
    neighbor = []
    neighbor_c = []
    alpha_c_idx = alpha_c_index_by_category(smile, category0)[0]
    # 邻位必须在环上
    for nei in mol.GetAtomWithIdx(alpha_c_idx).GetNeighbors():
        for ri in ring:
            if nei.GetIdx() in ri:
                neighbor.append(nei.GetIdx())
    # 邻位必须为c
    for n in neighbor:
        if mol.GetAtomWithIdx(n).GetAtomicNum() == 6 or mol.GetAtomWithIdx(n).GetAtomicNum() == 7 \
                or mol.GetAtomWithIdx(n).GetAtomicNum() == 8 or mol.GetAtomWithIdx(n).GetAtomicNum() == 16:
            neighbor_c.append(n)

    return neighbor, neighbor_c


# alpha-c的化学键类型，是否含有DOUBLE，TRIPLE
def alpha_c_bonds_type(smile, category0):
    flag = False
    mol = Chem.MolFromSmiles(smile)
    # alpha-c列表
    alpha_c_idx_list = alpha_c_index_by_category(smile, category0)
    for alpha_c_idx in alpha_c_idx_list:
        # 获取alpha-c周围的原子
        neighbors = mol.GetAtomWithIdx(alpha_c_idx).GetNeighbors()
        for neighbor in neighbors:
            # 获取相邻两原子的化学键类型
            bond_type = mol.GetBondBetweenAtoms(alpha_c_idx, neighbor.GetIdx()).GetBondType()
            # alpha-c含有双键或三键，符合烯类
            if str(bond_type) in ['DOUBLE', 'TRIPLE']:
                flag = True
                return flag
    return flag


# 自定义芳香性环的判断
def is_ring_custom_ar(smile, category0):
    # 如果满足rdkit定义的芳香性，则为芳香
    if is_ring_aromatic(smile, category0):
        return 1
    # 未满足rdkit定义的，则用原子连接的化学键个数判断
    else:
        # alpha-c列表
        alpha_c_idx_list = alpha_c_index_by_category(smile, category0)
        # 环列表
        mol = Chem.MolFromSmiles(smile)
        ring_atom_idx_list = mol.GetRingInfo().AtomRings()
        # 找出含有alpha-c的环
        alpha_c_rings = []
        for alpha_c_idx in alpha_c_idx_list:
            for ring_atom_idx in ring_atom_idx_list:
                if alpha_c_idx in ring_atom_idx:
                    alpha_c_rings.append(ring_atom_idx)
        sp2_flag = 1
        if len(alpha_c_rings):
            for alpha_c_ring in alpha_c_rings:
                for idx in alpha_c_ring:
                    if str(mol.GetAtomWithIdx(idx).GetHybridization()) != 'SP2':
                        sp2_flag = 0
        else:
            sp2_flag = 0
        return sp2_flag


# 解决任意层次的嵌套元素提取
def get_item(total: Iterable) -> list:
    # 创建用于计算的闭包
    def calculate(lst: Iterable):  # 传入一个可迭代对象
        for item in lst:
            # 如果其中元素不可迭代，说明到达嵌套列表底层。将这个元素加到储存最终结果的result列表中，用return结束此次递归
            if not isinstance(item, Iterable):  # 这里判断元素是否可以迭代
                result.append(item)
            # 如果元素依旧可以迭代，调用递归对这个元素进行计算
            else:
                calculate(item)

    result = []  # 创建储存结果的列表
    calculate(total)  # 调用闭包计算
    return result  # 返回最终结果


# 判断alpha-c list与成环原子索引的关系，优先考虑了环上的alpha-c
def unique_alpha_c_idx_in_ring(alpha_c_list, smile):
    m = Chem.MolFromSmiles(smile)
    # 环原子索引
    ring = m.GetRingInfo().AtomRings()
    deduplicated = []
    # 1个alpha-c list与不同环做交集
    for ri in ring:
        i = set(alpha_c_list).intersection(set(ri))
        # 如果有交集，只取一个
        if len(i):
            deduplicated.append(list(i)[0])
    # 一直无交集，则保持原样
    if not len(deduplicated):
        deduplicated = alpha_c_list

    return deduplicated


# 删除某列表中的所有特定元素
def delete_all_x_idx(li, x):
    while True:
        if x in li:
            li.remove(x)
            continue
        else:
            break
    return li


# if __name__ == '__main__':
#     print(is_ring_custom_ar('O=C1COCc2ccccc12', 'Ketone'))
# print(identify_functional_group('C(=O)(CO)O', 'C(=O)O')[0][0][0])
# print(alpha_c_index_by_category('C(=O)(CO)O', 'Alcohol'))
# print(count_H_on_alpha_c('C#CCC(CC#CC)(C(=O)OC)C(=O)OC', 'Alkyne'))
# print(alpha_c_index_by_category('CCC(C)=O', 'Ketone'))
# print(alpha_c_index_by_category('C#CCC(CC#CC)(C(=O)OC)C(=O)OC', 'Alkyne'))
# print(identify_functional_group('O=C(O)C1CCC1', C0FG.Carboxylic_Acid.value))
# print(is_free_H('OB(O)c1cn[nH]c1', 'Boronic Acid'))
# print(is_free_H('O=C(O)C1CCO1', "Carboxylic Acid"))
# print(is_free_H('OB(O)c1cc(F)ccc1C(F)(F)F', "Boronic Acid"))
# print(is_alpha_ring_heterocyclic('CC1(C)OB(OC1(C)C)c1ccc(nc1)c1ccccc1', "Boronic Acid"))
# print(count_H('C1=CNCC1', 3))
# print(neighbor_c_idx_of_alpha_c('CN1CCN(c2ccc(B3OC(C)(C)C(C)(C)O3)cc2)CC1', 'Boronic Acid'))
# print(neighbor_c_idx_of_alpha_c('O=C(O)c1ccno1', 'Carboxylic Acid'))
# print(is_idx_in_double_ring('Oc1nccc2c(Br)[nH]nc12', 9))



# -*- coding: utf-8 -*-
# @Time : 2021/11/24 17:15
# @Author : EdwardTsai
# @Email : caiweimin@megarobo.tech
# @File : result.py
# @Project : chem-mesh-notebook
# @Introduction : For PyPI. You can get classification result directly through the method.
from classification.rules import classify


def custom_classification(smiles, category0):
    result = classify(smiles, category0)
    return result.toJson()



# -*- coding: utf-8 -*-
# @Time : 2021/11/24 17:30
# @Author : EdwardTsai
# @Email : caiweimin@megarobo.tech
# @File : setup.py
# @Project : chem-mesh-notebook
# @Introduction :
import os
import setuptools

# 如果readme文件中有中文，那么这里要指定encoding='utf-8'，否则会出现编码错误
with open(os.path.join(os.path.dirname(__file__), 'README.md'), encoding='utf-8') as readme:
    README = readme.read()

# 允许setup.py在任何路径下执行
os.chdir(os.path.normpath(os.path.join(os.path.abspath(__file__), os.pardir)))

setuptools.setup(
    name="classification-mega-chem-mesh",  # 库名，需要在pypi中唯一
    version="0.0.24",  # 版本号
    author="Jinrui.Duan & Weimin.Cai",  # 作者
    author_email="caiweimin@megarobo.tech",  # 作者邮箱（方便使用者发现问题后联系我们）
    description="A Compound classification engine",  # 简介
    long_description="This is a private tool for compound classification",  # 详细描述（一般会写在README.md中）
    long_description_content_type="text/markdown",  # README.md中描述的语法（一般为markdown）
    url="https://gitee.com/chenmesh/chem-mesh-notebook",  # 库/项目主页，一般我们把项目托管在GitHub，放该项目的GitHub地址即可
    packages=setuptools.find_packages(),  # 默认值即可，这个是方便以后我们给库拓展新功能的
    classifiers=[  # 指定该库依赖的Python版本、license、操作系统之类的
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    install_requires=[
        'numpy == 1.21.3',
        'rdkit-pypi == 2021.9.2',
    ],
    python_requires='>=3.6',
)



import warnings
from rdkit import Chem
from classification.models import Category0FunctionalGroup as C0FG
from classification.models import Category0, Category1, Category2, Category2Aromatic, \
    Category2Aliphatic, Category3, Category4, Category5, CategoryResult, Category4Smarts, Category4FiveMembered, \
    Category4SixMembered, Hydrolysis, Steric_Hindrance
from classification.utils import identify_functional_group as ifg
from classification.utils import functional_group_from_category0 as fgfc0
from classification.utils import convert_category_0, match_alpha_c_ali_aro, is_free_H, is_fused_ring, is_ring_x_atom, \
    alpha_c_index_by_category, count_H, count_ring_atom, locate_alpha_c_index_on_heterocyclic, \
    mapping_from_position_list, is_hydrolysis, neighbor_c_idx_of_alpha_c, is_ring_aromatic, unique_alpha_c_idx_in_ring, \
    is_idx_in_ring_and_aromatic, functional_group_from_category0, count_H_on_alpha_c, alpha_c_bonds_type, \
    is_ring_custom_ar, is_alpha_ring_heterocyclic, is_idx_in_double_ring
from typing import List


# 根据输入的SMILES判断其官能团，可能有一个或多个
def get_category_0(smile: str) -> List[str]:
    category_0_list = []
    # verify Halide
    if ifg(smile, C0FG.Halide.value)[1]:
        category_0_list.append(Category0.Halide.value)
    # verify Amine
    if ifg(smile, C0FG.Amine.value)[1]:
        category_0_list.append(Category0.Amine.value)
    # verify Alcohol
    if ifg(smile, C0FG.Alcohol.value)[1]:
        category_0_list.append(Category0.Alcohol.value)
    # verify Boronic Acid
    if ifg(smile, C0FG.Boronic_Acid.value)[1]:
        category_0_list.append(Category0.Boronic_Acid.value)
    # verify Carboxylic Acid
    if ifg(smile, C0FG.Carboxylic_Acid.value)[1]:
        category_0_list.append(Category0.Carboxylic_Acid.value)
    # verify Acid Chloride
    if ifg(smile, C0FG.Acid_Chloride.value)[1]:
        category_0_list.append(Category0.Acid_Chloride.value)
    # verify Sulfonyl Chloride
    if ifg(smile, C0FG.Sulfonyl_Chloride.value)[1]:
        category_0_list.append(Category0.Sulfonyl_Chloride.value)
    # verify Aldehyde
    if ifg(smile, C0FG.Aldehyde.value)[1]:
        category_0_list.append(Category0.Aldehyde.value)
    # verify Ketone
    if ifg(smile, C0FG.Ketone.value)[1]:
        category_0_list.append(Category0.Ketone.value)
    # # verify Alkyne
    # if ifg(smile, C0FG.Alkyne.value)[1]:
    #     category_0_list.append(Category0.Alkyne.value)
    # # verify Olefin
    # if ifg(smile, C0FG.Olefin.value)[1]:
    #     category_0_list.append(Category0.Olefin.value)
    # # verify Epoxides
    # if ifg(smile, C0FG.Epoxides.value)[1]:
    #     category_0_list.append(Category0.Epoxides.value)
    # # verify Hydrazines
    # if ifg(smile, C0FG.Hydrazines.value)[1]:
    #     category_0_list.append(Category0.Hydrazines.value)
    # # verify Isocyanates
    # if ifg(smile, C0FG.Isocyanates.value)[1]:
    #     category_0_list.append(Category0.Isocyanates.value)
    # # verify Isothiocyanates
    # if ifg(smile, C0FG.Isothiocyanates.value)[1]:
    #     category_0_list.append(Category0.Isothiocyanates.value)
    # # verify Nitriles
    # if ifg(smile, C0FG.Nitriles.value)[1]:
    #     category_0_list.append(Category0.Nitriles.value)
    # # verify Ester
    # if ifg(smile, C0FG.Ester1.value)[1]:
    #     category_0_list.append(Category0.Ester.value)
    # # verify Ester
    # if ifg(smile, C0FG.Ester2.value)[1]:
    #     category_0_list.append(Category0.Ester.value)
    # # verify Ester
    # if ifg(smile, C0FG.Ester3.value)[1]:
    #     category_0_list.append(Category0.Ester.value)
    # # verify Ester
    # if ifg(smile, C0FG.Ester4.value)[1]:
    #     category_0_list.append(Category0.Ester.value)
    # # verify Ester
    # if ifg(smile, C0FG.Ester5.value)[1]:
    #     category_0_list.append(Category0.Ester.value)
    # # verify Ester
    # if ifg(smile, C0FG.Ester6.value)[1]:
    #     category_0_list.append(Category0.Ester.value)

    return category_0_list


def get_category_1(smile: str, category0: str) -> Category1:
    if is_free_H(smile, category0):
        return Category1.Free_Hydrogen
    else:
        if category0 == Category0.Halide.value:
            ali, aro = match_alpha_c_ali_aro(smile, fgfc0(smile, category0)[1])
            # if ifg(smile, 'C(F)')[1]:
            #     if len(aro) > 0:
            #         return Category1.Aromatic
            #     else:
            #         return Category1.Aliphatic
            # else:
            if len(ali) > 0:
                return Category1.Aliphatic
            else:
                return Category1.Aromatic
        elif category0 == Category0.Alcohol.value:
            if alpha_c_bonds_type(smile, category0):
                # 如果是烯醇，再判断一次自定义的芳香性，再给出芳香or烯醇的结论
                if is_ring_custom_ar(smile, category0):
                    return Category1.Aromatic
                else:
                    return Category1.Enol
            else:
                if len(match_alpha_c_ali_aro(smile, fgfc0(smile, category0)[1])[1]):
                    if is_ring_aromatic(smile, category0):
                        return Category1.Aromatic
                    else:
                        return Category1.Aliphatic
                else:
                    return Category1.Aliphatic
        else:
            if len(match_alpha_c_ali_aro(smile, fgfc0(smile, category0)[1])[1]):
                if is_ring_aromatic(smile, category0):
                    return Category1.Aromatic
                else:
                    return Category1.Aliphatic
            else:
                return Category1.Aliphatic


def get_category_exclude_freeH(smile: str, category0: str) -> Category1:
    if category0 == Category0.Halide.value:
        ali, aro = match_alpha_c_ali_aro(smile, fgfc0(smile, category0)[1])
        if len(ali) > 0:
            return Category1.Aliphatic
        else:
            return Category1.Aromatic
    elif category0 == Category0.Alcohol.value:
        if alpha_c_bonds_type(smile, category0):
            # 如果是烯醇，再判断一次自定义的芳香性，再给出芳香or烯醇的结论
            if is_ring_custom_ar(smile, category0):
                return Category1.Aromatic
            else:
                return Category1.Enol
        else:
            if len(match_alpha_c_ali_aro(smile, fgfc0(smile, category0)[1])[1]):
                if is_ring_aromatic(smile, category0):
                    return Category1.Aromatic
                else:
                    return Category1.Aliphatic
            else:
                return Category1.Aliphatic
    else:
        if len(match_alpha_c_ali_aro(smile, fgfc0(smile, category0)[1])[1]):
            if is_ring_aromatic(smile, category0):
                return Category1.Aromatic
            else:
                return Category1.Aliphatic
        else:
            return Category1.Aliphatic


def get_category_2(smile: str, category0: str):
    cate1_str = get_category_1(smile, category0).value
    # 父分类为芳香类
    if cate1_str == Category1.Aromatic.value:
        # 先判断并环
        if is_fused_ring(smile, category0):
            return Category2Aromatic.Fused_Ring
        else:
            # 芳香类问题排查，不能使用键判断，从alpha-c索引入手查是否包含在环索引元组中，然后根据个数判断5元6元，利用子结构判断是否苯环
            ring_len = is_ring_x_atom(smile, category0)
            if len(ring_len):
                if ring_len[0] == 6 and ifg(smile, 'O=C1C=C(O)C(=O)cc1')[1] or ifg(smile, 'O=C1C=CC(=O)C=C1')[1]:
                    return Category2Aromatic.Quinone
                elif ring_len[0] == 6 and not is_alpha_ring_heterocyclic(smile, category0):
                    return Category2Aromatic.Benzene
                elif ring_len[0] == 5 or ring_len[0] == 6 and is_alpha_ring_heterocyclic(smile, category0):
                    return Category2Aromatic.Heterocyclic
                else:
                    return Category2.Error
            else:
                return Category2.Error

    # 父分类为脂肪类
    elif cate1_str == Category1.Aliphatic.value:
        # 如果是胺类，则判断N原子的氢原子数
        if category0 == Category0.Amine.value:
            n_h_num = []
            res = fgfc0(smile, category0)
            for match in res[0]:
                h_num = count_H(smile, match[0])
                n_h_num.append(h_num)
            if 2 in n_h_num:
                if ifg(smile, '[R]C[NH2]')[1]:
                    if is_idx_in_ring_and_aromatic(smile, ifg(smile, '[R]C[NH2]')[0][0][0]):
                        return Category2Aliphatic.Benzyl
                    else:
                        return Category2Aliphatic.Primary
                else:
                    return Category2Aliphatic.Primary
            elif 3 in n_h_num:
                if ifg(smile, '[R]C[NH3+]')[1]:
                    if is_idx_in_ring_and_aromatic(smile, ifg(smile, '[R]C[NH3+]')[0][0][0]):
                        return Category2Aliphatic.Benzyl
                    else:
                        return Category2Aliphatic.Primary
                else:
                    return Category2Aliphatic.Primary
            elif 1 in n_h_num:
                return Category2Aliphatic.Secondary
            elif 0 in n_h_num:
                return Category2Aliphatic.Tertiary
            else:
                return Category2.Error

        # 醇类和醛类先判断beta位连接的是否为芳香环，其次再判断alpha-c上H的个数分
        elif category0 == Category0.Alcohol.value or category0 == Category0.Aldehyde.value:
            fg = functional_group_from_category0(smile, category0)[1]
            if ifg(smile, '[R]C'+fg)[1]:
                if is_idx_in_ring_and_aromatic(smile, ifg(smile, '[R]C'+fg)[0][0][0]):
                    return Category2Aliphatic.Benzyl
                else:
                    return count_H_on_alpha_c(smile, category0)
            else:
                return count_H_on_alpha_c(smile, category0)

        # 其他大类类根据alpha-c上H的个数分
        else:
            return count_H_on_alpha_c(smile, category0)
    # 父分类为活泼氢，不做处理
    else:
        return Category2.Default


# 上一级分类为杂环
def get_category_3(smile: str, category0: str) -> Category3:
    cate2_str = get_category_2(smile, category0).value
    if cate2_str == Category2Aromatic.Heterocyclic.value:
        alpha_c_idx_list = alpha_c_index_by_category(smile, category0)
        u_alpha_c = unique_alpha_c_idx_in_ring(alpha_c_idx_list, smile)
        ring_atom_num = []
        for u in u_alpha_c:
            num = count_ring_atom(smile, u)
            ring_atom_num.append(num)
        # 考虑了官能团连接一个或多个5或6元环，同时连接5且6元环的情况不在该范围中
        if sum(ring_atom_num) == 5 * len(ring_atom_num):
            return Category3.FiveMembered
        elif sum(ring_atom_num) == 6 * len(ring_atom_num):
            return Category3.SixMembered
        # 如果集合[5,6]，又因为杂环大于苯环，有5出现必定是杂环
        elif 5 in ring_atom_num:
            return Category3.FiveMembered
        else:
            return Category3.Error

    # 父分类不是杂环，不做处理
    else:
        return Category3.Default


# 上一分类为5元或6元杂环
def get_category_4(smile: str, category0: str):
    cate3_str = get_category_3(smile, category0).value
    if cate3_str == Category3.FiveMembered.value:
        if ifg(smile, Category4Smarts.Pyrrole_3H.value)[1]:
            return Category4FiveMembered.Pyrrole_3H
        elif ifg(smile, Category4Smarts.Pyrrole_2H.value)[1]:
            return Category4FiveMembered.Pyrrole_2H
        elif ifg(smile, Category4Smarts.Pyrrole_2H_Enol.value)[1]:
            return Category4FiveMembered.Pyrrole_2H
        elif ifg(smile, Category4Smarts.Pyrazole_3H.value)[1]:
            return Category4FiveMembered.Pyrazole_3H
        elif ifg(smile, Category4Smarts.Imidazole_4H.value)[1]:
            return Category4FiveMembered.Imidazole_4H
        elif ifg(smile, Category4Smarts.Triazole_124_3H.value)[1]:
            return Category4FiveMembered.Triazole_124_3H
        elif ifg(smile, Category4Smarts.Triazole_123_4H.value)[1]:
            return Category4FiveMembered.Triazole_123_4H
        elif ifg(smile, Category4Smarts.Tetrazole_5H.value)[1]:
            return Category4FiveMembered.Tetrazole_5H
        elif ifg(smile, Category4Smarts.Furan.value)[1]:
            return Category4FiveMembered.Furan
        elif ifg(smile, Category4Smarts.Oxazole.value)[1]:
            return Category4FiveMembered.Oxazole
        elif ifg(smile, Category4Smarts.Isoxazole.value)[1]:
            return Category4FiveMembered.Isoxazole
        elif ifg(smile, Category4Smarts.Oxadiazole_123.value)[1]:
            return Category4FiveMembered.Oxadiazole_123
        elif ifg(smile, Category4Smarts.Oxadiazole_124.value)[1]:
            return Category4FiveMembered.Oxadiazole_124
        elif ifg(smile, Category4Smarts.Oxatriazole_1235.value)[1]:
            return Category4FiveMembered.Oxatriazole_1235
        elif ifg(smile, Category4Smarts.Oxatriazole_1234.value)[1]:
            return Category4FiveMembered.Oxatriazole_1234
        elif ifg(smile, Category4Smarts.Tetranitrogen_monooxide.value)[1]:
            return Category4FiveMembered.Tetranitrogen_monooxide
        elif ifg(smile, Category4Smarts.Thiophene.value)[1]:
            return Category4FiveMembered.Thiophene
        elif ifg(smile, Category4Smarts.Thiazole.value)[1]:
            return Category4FiveMembered.Thiazole
        elif ifg(smile, Category4Smarts.Isothiazole.value)[1]:
            return Category4FiveMembered.Isothiazole
        elif ifg(smile, Category4Smarts.Thiadiazole_123.value)[1]:
            return Category4FiveMembered.Thiadiazole_123
        elif ifg(smile, Category4Smarts.Thiadiazole_124.value)[1]:
            return Category4FiveMembered.Thiadiazole_124
        elif ifg(smile, Category4Smarts.Thiadiazole_134.value)[1]:
            return Category4FiveMembered.Thiadiazole_134
        elif ifg(smile, Category4Smarts.Thiatriazole_1235.value)[1]:
            return Category4FiveMembered.Thiatriazole_1235
        elif ifg(smile, Category4Smarts.Thiatriazole_1234.value)[1]:
            return Category4FiveMembered.Thiatriazole_1234
        elif ifg(smile, Category4Smarts.Tetranitrogen_monosulfide.value)[1]:
            return Category4FiveMembered.Tetranitrogen_monosulfide
        elif ifg(smile, Category4Smarts.Pyrrole.value)[1]:
            return Category4FiveMembered.Pyrrole
        elif ifg(smile, Category4Smarts.Imidazole_nH.value)[1]:
            return Category4FiveMembered.Imidazole
        elif ifg(smile, Category4Smarts.Imidazole_n.value)[1]:
            return Category4FiveMembered.Imidazole
        elif ifg(smile, Category4Smarts.Pyrazole.value)[1]:
            return Category4FiveMembered.Pyrazole
        elif ifg(smile, Category4Smarts.Triazole_123.value)[1]:
            return Category4FiveMembered.Triazole_123
        elif ifg(smile, Category4Smarts.Triazole_124.value)[1]:
            return Category4FiveMembered.Triazole_124
        elif ifg(smile, Category4Smarts.Tetrazole_1235.value)[1]:
            return Category4FiveMembered.Tetrazole
        elif ifg(smile, Category4Smarts.Tetrazole_1234.value)[1]:
            return Category4FiveMembered.Tetrazole
        elif ifg(smile, Category4Smarts.Pentazole_1H.value)[1]:
            return Category4FiveMembered.Pentazole_1H
        else:
            return Category4.Error
    elif cate3_str == Category3.SixMembered.value:
        if ifg(smile, Category4Smarts.Pyridine.value)[1]:
            return Category4SixMembered.Pyridine
        elif ifg(smile, Category4Smarts.Pyridazine.value)[1]:
            return Category4SixMembered.Pyridazine
        elif ifg(smile, Category4Smarts.Pyrimidine.value)[1]:
            return Category4SixMembered.Pyrimidine
        elif ifg(smile, Category4Smarts.Pyrazine.value)[1]:
            return Category4SixMembered.Pyrazine
        elif ifg(smile, Category4Smarts.Triazine_123.value)[1]:
            return Category4SixMembered.Triazine_123
        elif ifg(smile, Category4Smarts.Triazine_124.value)[1]:
            return Category4SixMembered.Triazine_124
        elif ifg(smile, Category4Smarts.Triazine_135.value)[1]:
            return Category4SixMembered.Triazine_135
        elif ifg(smile, Category4Smarts.Tetrazine_1234.value)[1]:
            return Category4SixMembered.Tetrazine_1234
        elif ifg(smile, Category4Smarts.Tetrazine_1245.value)[1]:
            return Category4SixMembered.Tetrazine_1245
        elif ifg(smile, Category4Smarts.Tetrazine_1235.value)[1]:
            return Category4SixMembered.Tetrazine_1235
        elif ifg(smile, Category4Smarts.Pentazine.value)[1]:
            return Category4SixMembered.Pentazine
        elif ifg(smile, Category4Smarts.Hexazine.value)[1]:
            return Category4SixMembered.Hexazine
        else:
            return Category4.Error
    else:
        return Category4.Error


def get_category_5(smile: str, category0: str) -> List:
    cate4_str = get_category_4(smile, category0).value
    if cate4_str == Category4FiveMembered.Pyrrole_3H.value:
        pos = locate_alpha_c_index_on_heterocyclic(smile, category0, Category4Smarts.Pyrrole_3H.value)
        return mapping_from_position_list(pos)

    elif cate4_str == Category4FiveMembered.Pyrrole_2H.value:
        pos = locate_alpha_c_index_on_heterocyclic(smile, category0, Category4Smarts.Pyrrole_2H.value)
        if len(pos):
            return mapping_from_position_list(pos)
        else:
            pos = locate_alpha_c_index_on_heterocyclic(smile, category0, Category4Smarts.Pyrrole_2H_Enol.value)
            return mapping_from_position_list(pos)

    elif cate4_str == Category4FiveMembered.Pyrazole_3H.value:
        pos = locate_alpha_c_index_on_heterocyclic(smile, category0, Category4Smarts.Pyrazole_3H.value)
        return mapping_from_position_list(pos)

    elif cate4_str == Category4FiveMembered.Imidazole_4H.value:
        pos = locate_alpha_c_index_on_heterocyclic(smile, category0, Category4Smarts.Imidazole_4H.value)
        return mapping_from_position_list(pos)

    elif cate4_str == Category4FiveMembered.Triazole_124_3H.value:
        pos = locate_alpha_c_index_on_heterocyclic(smile, category0, Category4Smarts.Triazole_124_3H.value)
        return mapping_from_position_list(pos)

    elif cate4_str == Category4FiveMembered.Triazole_123_4H.value:
        pos = locate_alpha_c_index_on_heterocyclic(smile, category0, Category4Smarts.Triazole_123_4H.value)
        return mapping_from_position_list(pos)

    elif cate4_str == Category4FiveMembered.Tetrazole_5H.value:
        pos = locate_alpha_c_index_on_heterocyclic(smile, category0, Category4Smarts.Tetrazole_5H.value)
        return mapping_from_position_list(pos)

    elif cate4_str == Category4FiveMembered.Furan.value:
        pos = locate_alpha_c_index_on_heterocyclic(smile, category0, Category4Smarts.Furan.value)
        return mapping_from_position_list(pos)

    elif cate4_str == Category4FiveMembered.Oxazole.value:
        pos = locate_alpha_c_index_on_heterocyclic(smile, category0, Category4Smarts.Oxazole.value)
        return mapping_from_position_list(pos)

    elif cate4_str == Category4FiveMembered.Isoxazole.value:
        pos = locate_alpha_c_index_on_heterocyclic(smile, category0, Category4Smarts.Isoxazole.value)
        return mapping_from_position_list(pos)

    elif cate4_str == Category4FiveMembered.Oxadiazole_123.value:
        pos = locate_alpha_c_index_on_heterocyclic(smile, category0, Category4Smarts.Oxadiazole_123.value)
        return mapping_from_position_list(pos)

    elif cate4_str == Category4FiveMembered.Oxadiazole_124.value:
        pos = locate_alpha_c_index_on_heterocyclic(smile, category0, Category4Smarts.Oxadiazole_124.value)
        return mapping_from_position_list(pos)

    elif cate4_str == Category4FiveMembered.Oxatriazole_1235.value:
        pos = locate_alpha_c_index_on_heterocyclic(smile, category0, Category4Smarts.Oxatriazole_1235.value)
        return mapping_from_position_list(pos)

    elif cate4_str == Category4FiveMembered.Oxatriazole_1234.value:
        pos = locate_alpha_c_index_on_heterocyclic(smile, category0, Category4Smarts.Oxatriazole_1234.value)
        return mapping_from_position_list(pos)

    elif cate4_str == Category4FiveMembered.Tetranitrogen_monooxide.value:
        pos = locate_alpha_c_index_on_heterocyclic(smile, category0, Category4Smarts.Tetranitrogen_monooxide.value)
        return mapping_from_position_list(pos)

    elif cate4_str == Category4FiveMembered.Thiophene.value:
        pos = locate_alpha_c_index_on_heterocyclic(smile, category0, Category4Smarts.Thiophene.value)
        return mapping_from_position_list(pos)

    elif cate4_str == Category4FiveMembered.Thiazole.value:
        pos = locate_alpha_c_index_on_heterocyclic(smile, category0, Category4Smarts.Thiazole.value)
        return mapping_from_position_list(pos)

    elif cate4_str == Category4FiveMembered.Isothiazole.value:
        pos = locate_alpha_c_index_on_heterocyclic(smile, category0, Category4Smarts.Isothiazole.value)
        return mapping_from_position_list(pos)

    elif cate4_str == Category4FiveMembered.Thiadiazole_123.value:
        pos = locate_alpha_c_index_on_heterocyclic(smile, category0, Category4Smarts.Thiadiazole_123.value)
        return mapping_from_position_list(pos)

    elif cate4_str == Category4FiveMembered.Thiadiazole_124.value:
        pos = locate_alpha_c_index_on_heterocyclic(smile, category0, Category4Smarts.Thiadiazole_124.value)
        return mapping_from_position_list(pos)

    elif cate4_str == Category4FiveMembered.Thiadiazole_134.value:
        pos = locate_alpha_c_index_on_heterocyclic(smile, category0, Category4Smarts.Thiadiazole_134.value)
        return mapping_from_position_list(pos)

    elif cate4_str == Category4FiveMembered.Thiatriazole_1235.value:
        pos = locate_alpha_c_index_on_heterocyclic(smile, category0, Category4Smarts.Thiatriazole_1235.value)
        return mapping_from_position_list(pos)

    elif cate4_str == Category4FiveMembered.Thiatriazole_1234.value:
        pos = locate_alpha_c_index_on_heterocyclic(smile, category0, Category4Smarts.Thiatriazole_1234.value)
        return mapping_from_position_list(pos)

    elif cate4_str == Category4FiveMembered.Tetranitrogen_monosulfide.value:
        pos = locate_alpha_c_index_on_heterocyclic(smile, category0, Category4Smarts.Tetranitrogen_monosulfide.value)
        return mapping_from_position_list(pos)

    elif cate4_str == Category4FiveMembered.Pyrrole.value:
        pos = locate_alpha_c_index_on_heterocyclic(smile, category0, Category4Smarts.Pyrrole.value)
        return mapping_from_position_list(pos)

    elif cate4_str == Category4FiveMembered.Imidazole.value:
        pos = locate_alpha_c_index_on_heterocyclic(smile, category0, Category4Smarts.Imidazole_nH.value)
        if len(pos):
            return mapping_from_position_list(pos)
        else:
            pos = locate_alpha_c_index_on_heterocyclic(smile, category0, Category4Smarts.Imidazole_n.value)
            return mapping_from_position_list(pos)

    elif cate4_str == Category4FiveMembered.Pyrazole.value:
        pos = locate_alpha_c_index_on_heterocyclic(smile, category0, Category4Smarts.Pyrazole.value)
        return mapping_from_position_list(pos)

    elif cate4_str == Category4FiveMembered.Triazole_123.value:
        pos = locate_alpha_c_index_on_heterocyclic(smile, category0, Category4Smarts.Triazole_123.value)
        return mapping_from_position_list(pos)

    elif cate4_str == Category4FiveMembered.Triazole_124.value:
        pos = locate_alpha_c_index_on_heterocyclic(smile, category0, Category4Smarts.Triazole_124.value)
        return mapping_from_position_list(pos)

    elif cate4_str == Category4FiveMembered.Tetrazole.value:
        pos = locate_alpha_c_index_on_heterocyclic(smile, category0, Category4Smarts.Tetrazole_1235.value)
        if len(pos):
            return mapping_from_position_list(pos)
        else:
            pos = locate_alpha_c_index_on_heterocyclic(smile, category0, Category4Smarts.Tetrazole_1234.value)
            return mapping_from_position_list(pos)

    elif cate4_str == Category4FiveMembered.Pentazole_1H.value:
        pos = locate_alpha_c_index_on_heterocyclic(smile, category0, Category4Smarts.Pentazole_1H.value)
        return mapping_from_position_list(pos)

    elif cate4_str == Category4SixMembered.Pyridine.value:
        pos = locate_alpha_c_index_on_heterocyclic(smile, category0, Category4Smarts.Pyridine.value)
        return mapping_from_position_list(pos)

    elif cate4_str == Category4SixMembered.Pyridazine.value:
        pos = locate_alpha_c_index_on_heterocyclic(smile, category0, Category4Smarts.Pyridazine.value)
        return mapping_from_position_list(pos)

    elif cate4_str == Category4SixMembered.Pyrimidine.value:
        pos = locate_alpha_c_index_on_heterocyclic(smile, category0, Category4Smarts.Pyrimidine.value)
        return mapping_from_position_list(pos)

    elif cate4_str == Category4SixMembered.Pyrazine.value:
        pos = locate_alpha_c_index_on_heterocyclic(smile, category0, Category4Smarts.Pyrazine.value)
        return mapping_from_position_list(pos)

    elif cate4_str == Category4SixMembered.Triazine_123.value:
        pos = locate_alpha_c_index_on_heterocyclic(smile, category0, Category4Smarts.Triazine_123.value)
        return mapping_from_position_list(pos)

    elif cate4_str == Category4SixMembered.Triazine_124.value:
        pos = locate_alpha_c_index_on_heterocyclic(smile, category0, Category4Smarts.Triazine_124.value)
        return mapping_from_position_list(pos)

    elif cate4_str == Category4SixMembered.Triazine_135.value:
        pos = locate_alpha_c_index_on_heterocyclic(smile, category0, Category4Smarts.Triazine_135.value)
        return mapping_from_position_list(pos)

    elif cate4_str == Category4SixMembered.Tetrazine_1234.value:
        pos = locate_alpha_c_index_on_heterocyclic(smile, category0, Category4Smarts.Tetrazine_1234.value)
        return mapping_from_position_list(pos)

    elif cate4_str == Category4SixMembered.Tetrazine_1245.value:
        pos = locate_alpha_c_index_on_heterocyclic(smile, category0, Category4Smarts.Tetrazine_1245.value)
        return mapping_from_position_list(pos)

    elif cate4_str == Category4SixMembered.Tetrazine_1235.value:
        pos = locate_alpha_c_index_on_heterocyclic(smile, category0, Category4Smarts.Tetrazine_1235.value)
        return mapping_from_position_list(pos)

    elif cate4_str == Category4SixMembered.Pentazine.value:
        pos = locate_alpha_c_index_on_heterocyclic(smile, category0, Category4Smarts.Pentazine.value)
        return mapping_from_position_list(pos)

    elif cate4_str == Category4SixMembered.Hexazine.value:
        pos = locate_alpha_c_index_on_heterocyclic(smile, category0, Category4Smarts.Hexazine.value)
        return mapping_from_position_list(pos)

    else:
        return ["Category5 Error"]


def get_hydrolysis(smile, category0):
    if is_hydrolysis(smile, category0):
        return Hydrolysis.Yes
    else:
        return Hydrolysis.No


def get_steric_hindrance(smile, cate1, category0):
    mol = Chem.MolFromSmiles(smile)
    # 脂肪类，任意alpha-c上H个数小于2
    if cate1 == Category1.Aliphatic.value:
        alpha_c_idx = alpha_c_index_by_category(smile, category0)
        for ac in alpha_c_idx:
            if count_H(smile, ac) < 2:
                return Steric_Hindrance.Yes
        return Steric_Hindrance.No
    # 芳香类，苯环，仲胺也算有位阻
    # elif cate1 == Category1.Aromatic.value and cate2 == Category2Aromatic.Benzene.value and \
    #         ifg(smile, '[NH]c1ccccc1')[1]:
    #     return Steric_Hindrance.Yes
    # 芳香类，alpha-c邻位的c上H的个数为0
    elif cate1 == Category1.Aromatic.value:
        neighbor = neighbor_c_idx_of_alpha_c(smile, category0)[1]
        for n in neighbor:
            if not count_H(smile, n) and mol.GetAtomWithIdx(n).GetAtomicNum() == 6 \
                    and is_idx_in_double_ring(smile, n) != 2:
                return Steric_Hindrance.Yes
        return Steric_Hindrance.No
    # 活泼氢类，还要重新计算cate1，再按照脂肪和芳香计算
    elif cate1 == Category1.Free_Hydrogen.value:
        cate1_exh = get_category_exclude_freeH(smile, category0)
        if cate1_exh == Category1.Aliphatic.value:
            alpha_c_idx = alpha_c_index_by_category(smile, category0)
            for ac in alpha_c_idx:
                if count_H(smile, ac) < 2:
                    return Steric_Hindrance.Yes
            return Steric_Hindrance.No
        elif cate1_exh == Category1.Aromatic.value:
            neighbor = neighbor_c_idx_of_alpha_c(smile, category0)[1]
            for n in neighbor:
                if not count_H(smile, n) and mol.GetAtomWithIdx(n).GetAtomicNum() == 6 \
                        and is_idx_in_double_ring(smile, n) != 2:
                    return Steric_Hindrance.Yes
            return Steric_Hindrance.No
    else:
        return Steric_Hindrance.No


def classify(smile: str, category0: str) -> CategoryResult:
    category_0_list = get_category_0(smile)
    # 判断根据smile划分的大类是否与输入一致
    if category0 in category_0_list:
        # 包含关系，warning提示
        if len(category_0_list) > 1:
            warnings.warn('Compound belongs to multiple category_0!')
        # 无论包含还是等于，都进行下一步分类
        # 先判断是否易水解，与分类无关
        hydroly = get_hydrolysis(smile, category0)
        cate0 = convert_category_0(category0)
        cate1 = get_category_1(smile, category0)
        # cate1为芳香类
        if cate1.value == Category1.Aromatic.value:
            cate2 = get_category_2(smile, category0)
            sh = get_steric_hindrance(smile, cate1.value, category0)
            # cate2 为杂环
            if cate2.value == Category2Aromatic.Heterocyclic.value:
                cate3 = get_category_3(smile, category0)
                cate4 = get_category_4(smile, category0)
                cate5 = " ".join(get_category_5(smile, category0))
                return CategoryResult(cate0, cate1, cate2, cate3, cate4, cate5, hydroly, sh)
            else:
                return CategoryResult(cate0, cate1, cate2, Category3.Default, Category4.Default, "-", hydroly, sh)
        # cate1为脂肪类
        elif cate1.value == Category1.Aliphatic.value:
            cate2 = get_category_2(smile, category0)
            sh = get_steric_hindrance(smile, cate1.value, category0)
            return CategoryResult(cate0, cate1, cate2, Category3.Default, Category4.Default, "-", hydroly, sh)
        # cate1为活泼氢
        else:
            sh = get_steric_hindrance(smile, cate1.value, category0)
            return CategoryResult(cate0, cate1, Category2.Default, Category3.Default, Category4.Default, "-", hydroly,
                                  sh)
    elif category0 == 'Others':
        return CategoryResult(Category0.Others, Category1.Default, Category2.Default, Category3.Default,
                              Category4.Default, "-", Hydrolysis.Default, Steric_Hindrance.Default)

    # 判断有误，直接报错
    else:
        # return CategoryResult(Category0.Error, Category1.Default, Category2.Default, Category3.Default,
        #                       Category4.Default, "-", Hydrolysis.Default, Steric_Hindrance.Default)
        raise Exception('Illegal Category: {} - {}'.format(smile, category0))


# if __name__ == '__main__':
    # a = classify('CC(C)(C)OC(=O)N1CCN(S(=O)(=O)CCN)CC1', 'Amine')
    # a = classify('CC(C)(C)OC(=O)N1CCN(S(=O)(=O)CCN)CC1', 'Amine')  # BOC验证
    # a = classify('COC(=O)c1ccc(Cl)c(N)c1', 'Amine')  # 甲酯
    # a = classify('CCOC(=O)CC(C)(C)N', 'Amine')  # 乙酯
    # a = classify('COC(=O)CCCCCCN.Cl', 'Amine')  # 甲酯
    # a = classify('O=C(O)C1CCC1', 'Carboxylic Acid')
    # a = classify('OB(O)c1cn[nH]c1', 'Boronic Acid')  # 验证活泼氢
    # a = classify('CC1(C)OB(OC1(C)C)c1ccc(nc1)c1ccccc1', 'Boronic Acid')  # 验证吡啶变苯环
    # a = classify('Cn1cc(B2OC(C)(C)C(C)(C)O2)cn1', 'Boronic Acid')  # 验证吡啶变苯环(回归测试)
    # a = classify('CC1(C)OB(c2ccc(N3CCCCC3)nc2)OC1(C)C', 'Boronic Acid')  # 验证硼酸酯位阻
    # a = classify('O=C(O)c1ncon1', 'Carboxylic Acid')  # 验证羧酸位阻
    # a = classify('Cc1nn2c(O)nncc2c1Br', 'Halide')  # 验证卤代物位阻，有位阻
    # a = classify('c12c(cccc1c(c[nH]2)Br)N', 'Halide')  # 验证卤代物位阻，无位阻
    # a = classify('N#CCO', 'Alcohol')  # 验证位阻规则，alpha位小于2H
    # a = classify('CCC(C)=O', 'Ketone')  # 验证Ketone
    # a = classify('N#Cc1sc(Cl)nc1-c1ccc(Cl)cc1', 'Halide')  # 验证Halide
    # a = classify('[Cl-].[NH3+]C1CCCCC1(F)F', 'Amine')
    # a = classify('CC(=O)[O-]', 'Carboxylic Acid')
    # a = classify('O=C(O)CO', 'Alcohol')
    # a = classify('O=C1CCc2ccccc2CC1', 'Ketone')
    # a = classify('O=C(c1ccccc1)c2ccccn2', 'Ketone')
    # a = classify('O=C1COCc2ccccc12', 'Ketone')
    # a = classify('N#Cc1sc(Cl)nc1-c1ccc(Cl)cc1', 'Halide')
    # a = classify('O=C([O-])c1cccnc1.[Na+]', 'Carboxylic Acid')
    # a = classify('Cc1nnc(C(=O)[O-])s1.[Li+]', 'Carboxylic Acid')
    # print(a.category0)
    # print(a.category1)
    # print(a.category2)
    # print(a.category3)
    # print(a.category4)
    # print(a.category5)
    # print(a.hydrolysis)
    # print(a.steric_hindrance)
    # print(ifg('O=C1COCc2ccccc12', 'c1ccccc1'))
    # print(get_category_2('C#CCC(CC#CC)(C(=O)OC)C(=O)OC', 'Alkyne'))
    # print(get_category_2('CC1(C)OB(OC1(C)C)c1ccc(nc1)c1ccccc1', 'Boronic Acid'))
